import os
import json
import yaml


def find_file(path, file_name):
    for root, dirs, files in os.walk(path):
        if file_name in files:
            return os.path.join(root, file_name)


def find_dirs(path):
    results = []
    for root, dirs, files in os.walk(path):
        for d in dirs:
            results.append(d)
    return results


def list_file(path):
    files = []
    for file_name in sorted(os.listdir(path)):
        files.append(os.path.join(path, file_name))
    return files


def save_dict(path, target_dict):
    with open(path, 'w') as f:
        json.dump(target_dict, f)


def load_dict(path):
    with open(path) as f:
        return json.load(f)


def write_yaml(file_path, obj):
    with open(file_path, 'w') as f:
        yaml.dump(obj, f, allow_unicode=True, default_flow_style=False, sort_keys=False)


def load_yaml(file_path):
    with open(file_path, 'r') as f:
        data = yaml.safe_load(f)
    return data



import unittest
from BeautifulReport import BeautifulReport

if __name__ == "__main__":
    discover = unittest.defaultTestLoader.discover('.', pattern='*_test.py')
    print("Total Number Of Test Cases:", discover.countTestCases())
    result = BeautifulReport(discover)
    result.report(filename='test_report', report_dir="./target", description='Test Report', log_path=None)





# deduplicate rows
# change rows from
#   key1 key1-1 key1-1-1 value1
#   key1 key1-1 key1-1-2 value2
#   key1 key1-2 key1-2-1 value3
#   key1 key1-2 key1-2-2 value4
# to:
#   key1 key1-1 key1-1-1 value1
#               key1-1-2 value2
#        key1-2 key1-2-1 value1
#               key1-2-2 value2
def dedup_rows(rows, dedup_indexs):
    if len(rows) == 0:
        return rows
    rst = list()
    last_row = [''] * len(rows[0])
    for row in rows:
        deduped_row = list(row)
        for index, column_value in enumerate(row):
            if index in dedup_indexs and row[index] == last_row[index]:
                deduped_row[index] = ''
        rst.append(deduped_row)
        last_row = row
    return rst




import unittest
from pyutils.scheduling.event_driven import *


class ShareEvent(Event):
    def __init__(self, data):
        super().__init__(data)


class MarketEvent(Event):
    def __init__(self, data):
        super().__init__(data)


class ShareHandler(Handler):

    def __init__(self):
        self.data = list()

    def get_applicable_event_classes(self):
        return [ShareEvent]

    def get_data(self):
        return self.data

    def handle(self, event):
        self.data.append(event.get_data())


class ShareOrMarketHandler(ShareHandler):

    def get_applicable_event_classes(self):
        return [ShareEvent, MarketEvent]


class TestEventDriven(unittest.TestCase):

    def test_dispatch(self):
        dispatcher = Dispatcher()
        share_handler = ShareHandler()
        share_or_market_handler = ShareOrMarketHandler()

        # confirm that register non-event should raise error
        with self.assertRaises(RuntimeError) as err:
            dispatcher.register(ShareHandler, share_handler)
        self.assertTrue(err.exception.args[0].startswith("failed to register event type"))

        # confirm that register unsupported event should raise error
        with self.assertRaises(RuntimeError) as err:
            dispatcher.register(MarketEvent, share_handler)
        self.assertTrue(err.exception.args[0].startswith("unsupported event type"))

        # register events for handlers, then start dispatcher
        dispatcher.register(ShareEvent, share_handler)
        dispatcher.register(ShareEvent, share_or_market_handler)
        dispatcher.register(MarketEvent, share_or_market_handler)
        dispatcher.start()

        # dispatch two events, share event should be received by both handlers,
        # market event should only be received by share_or_market_handler.
        dispatcher.dispatch(ShareEvent("data1"))
        dispatcher.dispatch(MarketEvent("data2"))

        # stop dispatcher, and then waiting for all events to be processed
        dispatcher.stop()

        # confirm that share handler processed 1 event, another handler processed 2 events
        self.assertEqual(["data1"], share_handler.get_data())
        self.assertEqual(["data1", "data2"], share_or_market_handler.get_data())


if __name__ == '__main__':
    unittest.main()



import datetime


class Timer:

    def __init__(self, start=True):
        if start:
            self.start_time = datetime.datetime.now()

    def start(self):
        self.start_time = datetime.datetime.now()

    def get_elapse_ms(self):
        elapse = datetime.datetime.now() - self.start_time
        return elapse.total_seconds() * 1000

    def get_elapse_s(self):
        elapse = datetime.datetime.now() - self.start_time
        return elapse.total_seconds()




from abc import ABC, abstractmethod
from threading import Thread
import queue
import logging


class Dispatcher:

    def __init__(self, name="default"):
        self.__name = name
        self._mapping = dict()
        self._running = False
        self._thread = Thread(target=self.handle, args=())
        self._queue = queue.Queue()
        self._done = object()

    def start(self):
        self._running = True
        self._thread.start()
        logging.info("dispatcher {} started".format(self.__name))

    def stop(self):
        if self._running:
            self._queue.put(self._done)
            # Wait for actual termination
            self._thread.join()
            logging.info("dispatcher {} stopped".format(self.__name))
        else:
            logging.info("skip stopping dispatcher {} which is already stopped".format(self.__name))

    def register(self, event_class, handler):
        if not issubclass(event_class, Event):
            raise RuntimeError("[dispatcher {}] failed to register event type {0}".format(self.__name, event_class))
        if event_class not in handler.get_applicable_event_classes():
            raise RuntimeError("[dispatcher {}] unsupported event type {} for handler {}".format(
                self.__name, event_class, type(handler)))
        handlers = self._mapping.get(event_class)
        if handlers is None:
            handlers = [handler]
            self._mapping[event_class] = handlers
        else:
            handlers.append(handler)
        logging.info("[dispatcher {}] registered event_class {} with {} handlers: {}".format(
            self.__name, event_class, len(handlers), handlers))

    def unregister(self, event_class, handler):
        handlers = self._mapping.get(event_class)
        if handlers is not None:
            for existing_handler in handlers:
                if existing_handler == handler:
                    handlers.remove(existing_handler)
                    logging.info("[dispatcher {}] removed handler {} from event {}"
                                 .format(self.__name, existing_handler, event_class))

    def dispatch(self, event):
        self._queue.put(event)
        if self._queue.qsize() % 100 == 0:
            logging.info("[dispatcher {}] dispatch event {}, queue size: {}".format(
                self.__name, type(event), self._queue.qsize()))

    def handle(self):
        logging.info("[dispatcher {}] handle thread started!".format(self.__name))
        while self._running:
            event = self._queue.get()
            if event is self._done:
                break
            key = type(event)
            handlers = self._mapping.get(key)
            if handlers is None:
                raise RuntimeError("[dispatcher {}] failed to find handler for event type {}".format(
                    self.__name, type(event)))
            for handler in handlers:
                try:
                    handler.handle(event)
                except Exception as ex:
                    logging.error("[dispatcher {}] failed to handle event {}".format(self.__name, type(event)), ex)
            logging.debug("[dispatcher {}] processed event {}, queue size: {}".format(
                self.__name, event, self._queue.qsize()))
        logging.info("[dispatcher {}] handle thread stopped!!".format(self.__name))


class Event(ABC):

    def __init__(self, data=None):
        self.data = data

    def get_data(self):
        return self.data


class Handler(ABC):

    @abstractmethod
    def get_applicable_event_classes(self): pass

    @abstractmethod
    def handle(self, event): pass



from imapclient import IMAPClient


class MailReader:

    def __init__(self, imap_server, imap_port, user, pwd, charset='UTF-8', ssl=True):
        self.__imap_server = imap_server
        self.__imap_port = imap_port
        self.__user = user
        self.__pwd = pwd
        self.__charset = charset
        self.__ssl = ssl
        self.__client = None

    def start(self):
        self.__client = IMAPClient(self.__imap_server, ssl=self.__ssl, port=self.__imap_port)
        self.__client.login(self.__user, self.__pwd)
        self.__client.id_({"name": "IMAPClient", "version": "2.1.0"})

    def stop(self):
        self.__client.logout()

    # return mails and error if exception occurred
    def get_mails(self, folder, criteria, search_key_alias):
        mails = list()
        err = None
        try:
            self.__client.select_folder(folder)
            messages = self.__client.search(criteria=criteria, charset=self.__charset)
            if len(messages) > 0:
                for mail_id, data in self.__client.fetch(messages, search_key_alias.keys()).items():
                    mail = dict()
                    for key, alias in search_key_alias.items():
                        mail[alias] = data[bytes(key, self.__charset)].decode(self.__charset)
                    mails.append(mail)
        except Exception as ex:
            err = ex
        return mails, err



from apscheduler.schedulers.background import BackgroundScheduler
import time
import logging


class Scheduler:

    def __init__(self):
        self.scheduler = BackgroundScheduler()

    def start(self):
        self.scheduler.start()
        logging.info("scheduler started!")

    def stop(self):
        self.scheduler.remove_all_jobs()
        self.scheduler.shutdown()
        logging.info("scheduler stopped!")

    def add_periodical_job(self, job_id, func, interval_seconds, parameters=None):
        self.scheduler.add_job(func, 'interval', kwargs=parameters, seconds=interval_seconds, id=job_id)
        logging.info("added periodical job {} with interval {}s into scheduler".format(job_id, interval_seconds))

    def remove_job(self, job_id):
        self.scheduler.remove_job(job_id=job_id)
        logging.info("removed periodical job {} from scheduler".format(job_id))


if __name__ == "__main__":
    def test():
        print('test')
        # time.sleep(3)
    scheduler = Scheduler()
    scheduler.start()
    scheduler.add_periodical_job('test', test, 1)
    time.sleep(5)
    # scheduler.remove_job('test')
    # time.sleep(3)
    scheduler.stop()
    time.sleep(2)



import unittest
from pyutils.scheduling.scheduler import *


class TestScheduler(unittest.TestCase):

    def test_schedule(self):
        data1, data2 = list(), list()
        def append_data1():
            data1.append(1)
        def append_data2():
            data2.append(1)
        sched = Scheduler()
        sched.start()

        # add 2 periodical jobs
        sched.add_periodical_job("job1", append_data1, 0.01)
        sched.add_periodical_job("job2", append_data2, 0.01)

        # after 3 round, data controlled by jobs should be updated with 3 items
        time.sleep(0.033)
        self.assertEqual(3, len(data1))
        self.assertEqual(3, len(data2))

        # removed job1 should not continue counting
        sched.remove_job("job1")
        time.sleep(0.021)
        self.assertEqual(3, len(data1))
        self.assertEqual(5, len(data2))

        # all jobs are removed after scheduler stopped
        sched.stop()
        time.sleep(0.025)
        self.assertEqual(3, len(data1))
        self.assertEqual(5, len(data2))


if __name__ == '__main__':
    unittest.main()



from flask import Flask
from flask_mail import Mail, Message


class MailSender:

    def __init__(self, smtp_server, smtp_port, user, password, mail_recipients):
        self.__app = Flask(__name__)
        self.__app.config['MAIL_SERVER'] = smtp_server
        self.__app.config['MAIL_PORT'] = smtp_port
        self.__app.config['MAIL_USERNAME'] = user
        self.__app.config['MAIL_PASSWORD'] = password
        self.__app.config['MAIL_USE_TLS'] = False
        self.__app.config['MAIL_USE_SSL'] = True
        self.__mail_username = user
        self.__mail_recipients = mail_recipients
        self.__mail_client = Mail(self.__app)

    # return exception if occurred
    def send_mail(self, subject, body):
        try:
            msg = Message(
                sender=self.__mail_username,
                recipients=self.__mail_recipients,
                subject=subject,
                body=body)
            with self.__app.app_context():
                self.__mail_client.send(msg)
        except Exception as ex:
            return ex



from xml.dom import minidom


def read_conf(xml_file, xml_content=None):
    if xml_file is not None:
        doc = minidom.parse(xml_file)
    elif xml_content is not None:
        doc = minidom.parseString(xml_content)
    root = doc.documentElement
    conf_list = []
    for node in root.getElementsByTagName("property"):
        name = node.getElementsByTagName("name")[0].firstChild.data
        value_elem = node.getElementsByTagName("value")[0].firstChild
        value = value_elem.data if value_elem is not None else ''
        resource_elem = node.getElementsByTagName("source")
        resource = resource_elem[0].firstChild.data if resource_elem else ''
        conf_list.append({'key': name, 'value': value, 'resource': resource})
    return conf_list


def read_conf_2_dict(xml_file, xml_content=None):
    conf_list = read_conf(xml_file, xml_content)
    conf_dict = {}
    for conf_item in conf_list:
        conf_dict[conf_item['key']] = conf_item['value']
    return conf_dict



import time
from collections import OrderedDict
from pyutils.structures.time import TO_MINUTE_DATE_TIME_FORMAT


class QPSCalculator:

    def __init__(self):
        self.__qps_buckets_in_minutes = OrderedDict()

    def add(self, timestamp, num=1):
        struct_time = time.localtime(timestamp)
        minute_str = time.strftime(TO_MINUTE_DATE_TIME_FORMAT, struct_time)
        qps_buckets = self.__qps_buckets_in_minutes.get(minute_str)
        if qps_buckets is None:
            qps_buckets = [0] * 60
            self.__qps_buckets_in_minutes[minute_str] = qps_buckets
        sec_index = struct_time.tm_sec if struct_time.tm_sec < 60 else 59
        qps_buckets[sec_index] += num

    def get_qps_in_minutes(self):
        rst = OrderedDict()
        for minute_str, qps_buckets in self.__qps_buckets_in_minutes.items():
            rst[minute_str] = float(sum(qps_buckets)) / len(qps_buckets)
        return rst

    def get_qps_buckets_in_minutes(self):
        return self.__qps_buckets_in_minutes



from prettytable import PrettyTable
from textwrap import fill

TABLE_STYLE_COLUMN_MAX_WIDTH = 80


def generate_table(columns, rows):
    """
    :param columns: list of column names
    :param rows: list of rows
    :return: table
    """
    table = PrettyTable(field_names=tuple(columns), align='l')
    table.add_rows(rows)
    return table


def format_rows(rows, column_max_width=TABLE_STYLE_COLUMN_MAX_WIDTH):
    for row in rows:
        for index, row_item in enumerate(row):
            if isinstance(row_item, str) and len(row_item) > column_max_width:
                row[index] = fill(row_item, width=column_max_width)


def generate_grouped_info_table(grouped_keys, info_dict, format_fn_dict):
    """
    :param grouped_keys: list of grouped keys
    :param info_dict: info dict
    :param format_fn_dict: dict of format functions
    :return: table
    """
    rows = list()
    for grouped_key in grouped_keys:
        group_name = grouped_key.get('group_name')
        fields = grouped_key.get('fields')
        for no, field in enumerate(fields):
            key = field.get('key')
            value = info_dict.get(key)
            if value is None:
                continue
            format_fn = format_fn_dict.get(key) if format_fn_dict is not None else None
            if format_fn is not None:
                value = format_fn(value)
            alias = field.get('alias')
            row_data = [group_name if no == 0 else '', alias if alias is not None else key, value]
            rows.append(row_data)
    columns = ['Group', 'Field', 'Value']
    return generate_table(columns, rows)


def generate_compare_table(columns, row_keys, comparable_dicts, format_fn_dict):
    """
    :param columns: list of column names
    :param keys: list of keys
    :param comparable_dicts: list of comparable data dicts
    :param format_fn_dict: dict of format functions
    :return: table
    """
    rows = list()
    for row_key in row_keys:
        row = [row_key]
        format_fn = format_fn_dict.get(row_key) if format_fn_dict is not None else None
        for comparable_dict in comparable_dicts:
            value = comparable_dict.get(row_key)
            if value is None:
                row.append('')
            if format_fn is not None:
                value = format_fn(value)
            if isinstance(value, str) and len(value) > TABLE_STYLE_COLUMN_MAX_WIDTH:
                row.append(fill(value, width=TABLE_STYLE_COLUMN_MAX_WIDTH))
            else:
                row.append(value)
        rows.append(row)
    return generate_table(columns, rows)



import logging
import logging.config
import os
import yaml
from pyutils.structures import reflect as reflect_util
from pyutils.scheduling import event_driven
from pyutils.extender.constants import *
from pyutils.extender.framework import context, config, error


class Framework:

    def __init__(self, executing_path, conf_file, log_conf_file):
        self.__conf_file = conf_file
        self.__log_conf_file = log_conf_file
        self.__config = config.Config()
        self.__dispatcher = event_driven.Dispatcher()
        self.__context = context.Context(self.__config, executing_path, self.__dispatcher)

    def init(self):
        # init config
        self.__config.load_yaml(self.__conf_file)
        # init logging
        log_file_path = self.__config.get_or_default(DEFAULT_LOG_FILE_PATH, CONF_KEY_LOG_FILE_PATH)
        if not os.path.isabs(log_file_path):
            log_file_path = os.path.join(self.__context.get_executing_path(), log_file_path)
        init_logging(self.__context.get_executing_path(), self.__log_conf_file, log_file_path)
        # init extendable instances
        add_extendable_instances(self.__context, [CONF_KEY_PLUGINS, CONF_KEY_COMPONENTS], self.__context.add_component)
        add_extendable_instances(self.__context, [CONF_KEY_PLUGINS, CONF_KEY_SERVICES], self.__context.add_service)
        add_extendable_instances(self.__context, [CONF_KEY_PLUGINS, CONF_KEY_COMMANDS], self.__context.add_command)

    def get_context(self):
        return self.__context

    def start(self):
        # start dispatcher
        for service in self.__context.get_services().values():
            for event_clazz in service.get_applicable_event_classes():
                self.__dispatcher.register(event_clazz, service)
        self.__dispatcher.start()
        # start components and services in ascending order
        for key, component in self.__context.get_components().items():
            err = component.start()
            if err:
                raise RuntimeError("failed to start component {}: {}".format(key, str(err)))
        for service in self.__context.get_services().values():
            err = service.start()
            if err:
                err_msg = "failed to start service {}: {}".format(key, str(err))
                logging.error(err_msg)
                raise RuntimeError(err_msg)

    def stop(self):
        # stop services, components in descending order
        for service in reversed(self.__context.get_services().values()):
            service.stop()
        for component in reversed(self.__context.get_components().values()):
            component.stop()
        # stop dispatcher
        self.__dispatcher.stop()

    def execute_cmd(self, args):
        # run command
        commands = self.__context.get_commands()
        if len(args) == 0:
            logging.error("no command specified! valid commands: {}".format(list(commands.keys())))
            return EXIT_CODE_RUN_ARGUMENTS_ERROR
        command_name = args[0]
        command_args = args[1:]
        command = commands.get(command_name)
        if command is None:
            logging.error("command {} not found! valid commands: {}".format(command_name, list(commands.keys())))
            return EXIT_CODE_RUN_ARGUMENTS_ERROR
        try:
            output, err = command.run(command_args)
        except Exception as e:
            logging.debug("failed to run command {}".format(command_args), exc_info=e)
            err = e
        if err is not None:
            logging.error("failed to run command {}, error message: {}".format(command_args, str(err)))
            return EXIT_CODE_RUN_ARGUMENTS_ERROR if isinstance(err, error.ArgumentsError) else\
                EXIT_CODE_CONF_ERROR if isinstance(err, error.ConfError) else EXIT_CODE_UNKNOWN
        if isinstance(output, str):
            print(output)
        elif isinstance(output, list):
            for index, item in enumerate(output):
                print(item)
        return EXIT_CODE_NORMAL


def init_logging(executing_path, log_conf_file, log_file_path):
    if log_conf_file:
        with open(log_conf_file, 'r') as f_conf:
            dict_conf = yaml.load(f_conf, Loader=yaml.FullLoader)
        handlers = dict_conf.get('handlers')
        # initialize logs directory if necessary
        for handler in handlers.values():
            handler_class = handler.get('class')
            if handler_class.find('FileHandler') != -1:
                log_dir = os.path.dirname(log_file_path)
                if not os.path.isabs(log_dir):
                    log_dir = os.path.join(executing_path, log_dir)
                if not os.path.exists(log_dir):
                    os.makedirs(log_dir)
                handler['filename'] = os.path.join(log_dir, os.path.basename(log_file_path))
        logging.config.dictConfig(dict_conf)
    else:
        logging.info("logging to console since not found the log config file")


def init_context(executing_path, conf_file):
    ctx = context.Context(executing_path, conf_file)
    err = ctx.init()
    if err is not None:
        logging.error("{}. exit!".format(err))
        exit(EXIT_CODE_CONF_ERROR)
    return ctx


def add_extendable_instances(context, conf_keys, add_fn):
    conf = context.get_conf().get(*conf_keys)
    if not conf:
        logging.info("skip adding extendable instances of type {} without conf".format(type))
        return
    if isinstance(conf, list):
        conf_dict = dict()
        for conf_item in conf:
            key = conf_item.get(CONF_KEY_KEY)
            conf_dict[key] = conf_item
        conf = conf_dict
    for key, conf_item in conf.items():
        init_args = {CONF_KEY_KEY: key, CONF_KEY_CONTEXT: context}
        configured_init_args = conf_item.get(CONF_KEY_INIT_ARGS)
        if configured_init_args is not None:
            init_args.update(configured_init_args)
        try:
            construct_fn = reflect_util.reflect_fn(conf_item.get(CONF_KEY_CLASS))
            instance = construct_fn(**init_args)
        except Exception as ex:
            logging.error("failed to initialize instance: key={}, err={}".format(key, ex))
            exit(EXIT_CODE_CONF_ERROR)
        add_fn(key, instance)



import pandas
import numpy


def init():
    # show all columns
    pandas.set_option('display.max_columns', None)
    # show all rows
    pandas.set_option('display.max_rows', None)
    # keep row data in the same line
    pandas.set_option('display.width', None)
    # show all data
    pandas.set_option('max_colwidth', None)


def get_data_frame(dict_list_data, columns):
    nested_list_data = get_nested_list_data(dict_list_data, columns)
    return pandas.DataFrame(numpy.array(nested_list_data), columns=columns)


def get_nested_list_data(dict_list_data, columns):
    nested_list_data = list()
    for dict_obj in dict_list_data:
        sublist = list()
        for column in columns:
            sublist.append(dict_obj.get(column))
        nested_list_data.append(sublist)
    return nested_list_data





def add_tuple_by_indexes(tp1, tp2):
    return tuple(map(lambda x: x[0] + x[1], zip(tp1, tp2)))


def divide_tuple_by_indexes(tp1, tp2):
    return tuple(map(lambda x: round(x[0] / x[1], 3) if x[1] != 0 else 0.0, zip(tp1, tp2)))


def sub_tuple_by_indexes(tp1, tp2):
    return tuple(map(lambda x: round(x[0] - x[1], 3), zip(tp1, tp2)))


def min_tuple_by_indexes(tp1, tp2):
    return tuple(map(lambda x: min(x[0], x[1]), zip(tp1, tp2)))



import unittest
from pyutils.structures import iterable as iter_util


class TestIterable(unittest.TestCase):

    def test_has_negative(self):
        self.assertTrue(iter_util.has_negative((-1)))
        self.assertTrue(iter_util.has_negative((-5)))
        self.assertTrue(iter_util.has_negative((-1, 0)))
        self.assertTrue(iter_util.has_negative((0, -1)))
        self.assertTrue(iter_util.has_negative((-1, 1)))
        self.assertTrue(iter_util.has_negative((1, -1)))
        self.assertTrue(iter_util.has_negative((-3, 1)))
        self.assertTrue(iter_util.has_negative((-3, -1)))
        self.assertTrue(iter_util.has_negative((3, -3)))
        self.assertTrue(iter_util.has_negative((-1, 0, 1)))
        self.assertTrue(iter_util.has_negative((1, 2, -3)))

        self.assertFalse(iter_util.has_negative((0)))
        self.assertFalse(iter_util.has_negative((1)))
        self.assertFalse(iter_util.has_negative((6)))
        self.assertFalse(iter_util.has_negative((1, 0)))
        self.assertFalse(iter_util.has_negative((0, 1)))
        self.assertFalse(iter_util.has_negative((1, 1)))
        self.assertFalse(iter_util.has_negative((1, 2, 3)))
        self.assertFalse(iter_util.has_negative((1, 2, 3, 4, 5)))


if __name__ == '__main__':
    unittest.main()



import sys


class SimpleProgressBar:
    def __init__(self, description='', max=0, state=0, step=2):
        self.max = max
        self.state = state
        self.description = description
        self.step = step
        self.closed = False

    def __print(self, state):
        print("\r", end="")
        print("{}: {} {}".format(self.description, state, "/ {}".format(self.max) if self.max > 0 else ""),
              "▋" * (state // self.step), end="")

    def update(self, state=None):
        if state is not None:
            self.__print(state)
            if 0 < self.max == state:
                self.close()
            sys.stdout.flush()

    def close(self):
        if not self.closed:
            print("")
            self.closed = True



import os
from setuptools import setup, find_packages


def read(file_name):
    return open(os.path.join(os.path.dirname(__file__), file_name)).read()


setup(
    name="yolo-pyutils",
    version="0.1.17",
    author="Tao Yang",
    author_email="ytaofighting@gmail.com",
    description="yolo-pyutils provides utility functions for IO, data-processing, scheduling, common structures etc.",
    license="Apache License 2.0",
    keywords="utils",
    url="https://github.com/TaoYang526/yolo-pyutils",
    packages=find_packages(),
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Utilities",
        'License :: OSI Approved :: Apache Software License',
        "Programming Language :: Python :: 3"
    ],
)



import json
import time
import urllib
import urllib.request
import logging
from pyutils.processing.timer import Timer


class RestClient:

    def __init__(self, retry=3, retry_interval=0.1, timeout=0):
        self.__url = ''
        self.__urls = []
        self.__header = {}
        self.__timeout = timeout
        self.__retry = retry
        self.__retry_interval = retry_interval

    def url(self, url):
        self.__url = url
        self.__urls.clear()
        self.__urls.append(url)
        return self

    def urls(self, urls):
        self.__urls.clear()
        if type(urls) == list:
            self.__urls = urls
            self.__url = urls[0]
        elif type(urls) == str:
            self.__url = urls
            self.__urls.append(urls)
        return self

    def header(self, header):
        self.__header = header
        return self

    def timeout(self, timeout):
        self.__timeout = timeout
        return self

    def retry(self, retry):
        self.__retry = retry
        return self

    def switch_url(self):
        if len(self.__urls) > 1:
            for url in self.__urls:
                if url != self.__url:
                    self.__url = url
                    logging.info("switch url to %s", self.__url)
                    break
        return self

    def __build_request(self, data):
        if data is not None:
            if type(data) is dict:
                data = json.dumps(data)
            if self.__header is not None and len(self.__header) > 0:
                request = urllib.request.Request(self.__url, headers=self.__header, data=data)
            else:
                request = urllib.request.Request(self.__url, data=data)
        else:
            if self.__header is not None and len(self.__header) > 0:
                request = urllib.request.Request(self.__url, headers=self.__header)
            else:
                request = urllib.request.Request(self.__url)
        return request

    def __get_response(self, request):
        if self.__timeout > 0:
            response = urllib.request.urlopen(request, timeout=self.__timeout)
        else:
            response = urllib.request.urlopen(request)
        return response

    def __get_response_body(self, response):
        if response is None:
            return None
        response_body = response.read()
        if response_body is not None:
            response_body = response_body.decode()
        return response_body

    def __request(self, data, get_method=None):
        response = None
        retry = self.__retry
        timer = Timer(start=True)
        while retry > 0:
            try:
                request = self.__build_request(data)
                if get_method is not None:
                    request.get_method = lambda: get_method
                response = self.__get_response(request)
                break
            except Exception as ex:
                if isinstance(ex, urllib.error.HTTPError) and (ex.code == 404 or ex.code == 400):
                    raise ex
                retry -= 1
                logging.warning("Request %s error: %s, left retry cnt:%d", self.__url, str(ex), retry)
                if retry == 0:
                    raise ex
                self.switch_url()
                time.sleep(self.__retry_interval)
        logging.debug("requested url={}, get_method={}, data={}, elapse_ms={}, response_code={}, retry={}".format(
            self.__url, get_method, data, timer.get_elapse_ms(), response.getcode(), self.__retry - retry))
        return response

    def get_response_body_content(self):
        resp = self.__request(None)
        return self.__get_response_body(resp)

    def get(self):
        return self.__request(None)

    def post(self, data):
        return self.__request(data, 'POST')

    def put(self, data):
        return self.__request(data, 'PUT')

    def delete(self, data):
        return self.__request(data, 'DELETE')


def format_url(url, params):
    return format_urls(url, params)


def format_urls(urls, params):
    if params is not None and len(params) > 0:
        encoded_params = urllib.parse.urlencode(params)
        if type(urls) == list:
            new_urls = []
            for url in urls:
                url.append(__join_params(url, encoded_params))
            return new_urls
        elif type(urls) == str:
            return __join_params(urls, encoded_params)
    return urls


def __join_params(url, encoded_params):
    if url.find('?') == -1:
        return url + '?' + encoded_params
    else:
        return url + '&' + encoded_params


def append_url(url, appendix):
    return append_urls(url, appendix)


def append_urls(urls, appendix):
    if appendix is not None:
        if type(urls) == list:
            new_urls = []
            for url in urls:
                new_urls.append(url + appendix)
            return new_urls
        elif type(urls) == str:
            return urls + appendix
    return urls


def get_url_type(use_https):
    return 'https' if use_https is True else 'http'


def retrieve_file(remote_url, local_file_path, reporthook=None):
    urllib.request.urlretrieve(remote_url, filename=local_file_path, reporthook=reporthook)
